from datetime import datetime
import sys

def check_expiry():
    expiry_date = datetime(2025, 12, 31)
    if datetime.now() > expiry_date:
        print("This application has expired.")
        sys.exit(1)