import os
import re
from setuptools import setup, find_packages
from pathlib import Path

# Read the contents of your README file
this_directory = Path(__file__).parent
long_desc = (this_directory / "README.md").read_text()

def get_version():
    with open(os.path.join("lambend", "__version__.py")) as f:
        return re.search(r'__version__ = ["\'](.+?)["\']', f.read()).group(1)

setup(
    name='3pc-lambend',
    version=get_version(),
    description='Python library for analysis of shear deformable anisotropic laminates under transverse loading.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    author="Omprakash Seresta",
    author_email="oseresta@gmail.com",
    license="MIT",
    entry_points={
        'console_scripts': ['lambend = lambend.__main__:main']
    },
    packages=find_packages(),
    install_requires=[
        '3pc-material',
        '3pc-ply',
        '3pc-laminate',
        '3pc-panel',
        '3pc-loads',
        '3pc-utils',
        'numpy'
    ],
    keywords='laminate-bending composites transverse-loading shear-deformable anisotropic aerospace-engineering mechanics',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        'Blog': 'https://3pcomposites.com/blogs-1/f/3pcsolver002',
        'Analysis Tools': 'https://3pcomposites.com/analysis-tools',
        'Documentation': 'https://drive.google.com/drive/folders/1jDb-P5BOgJ-moVU20B7D8oVeHjHLAHTm',
    },
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False
)
